# Licensed under a BSD-style 3-clause license - see LICENSE.md.
# -*- coding: utf-8 -*-
"""
dlairflow
=========

Reusable code components for building `Apache Airflow®`_ DAGs.

.. _`Apache Airflow®`: https://airflow.apache.org
"""

__version__ = '0.1.5'
